import logging
import numpy as np
import collections
from p4p.client.thread import Context, Disconnected
from p4p.wrapper import Value
from .pva_codec import decompress
from pydm.data_plugins import is_read_only
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection
from pydm.widgets.channel import PyDMChannel
from qtpy.QtCore import QObject, Qt
from typing import Optional
import p4p

logger = logging.getLogger(__name__)


class Connection(PyDMConnection):

    def __init__(self, channel: PyDMChannel, address: str,
                 protocol: Optional[str] = None, parent: Optional[QObject] = None):
        """
        Manages the connection to a channel using the P4P library. A given channel can have multiple listeners.
        Parameters
        ----------
        channel : PyDMChannel
            The channel that this connection is connected to.
        address : str
             The address of the PV.
        protocol : str, optional
             The protocol prepended to the address.
        parent : QObject, optional
             The parent object of this connection.
        """
        super().__init__(channel, address, protocol, parent)
        self._connected = False
        self.nttable_data_location = PyDMPlugin.get_subfield(channel)
        self.monitor = P4PPlugin.context.monitor(name=self.address,
                                                 cb=self.send_new_value,
                                                 notify_disconnect=True)
        self.add_listener(channel)
        self._value = None
        self._severity = None
        self._precision = None
        self._enum_strs = None
        self._units = None
        self._upper_ctrl_limit = None
        self._lower_ctrl_limit = None
        self._upper_alarm_limit = None
        self._lower_alarm_limit = None
        self._upper_warning_limit = None
        self._lower_warning_limit = None
        self._timestamp = None

    def clear_cache(self) -> None:
        """ Clear out all the stored values of this connection. """
        self._value = None
        self._severity = None
        self._precision = None
        self._enum_strs = None
        self._units = None
        self._upper_ctrl_limit = None
        self._lower_ctrl_limit = None
        self._upper_alarm_limit = None
        self._lower_alarm_limit = None
        self._upper_warning_limit = None
        self._lower_warning_limit = None
        self._timestamp = None

    def send_new_value(self, value: Value) -> None:
        """ Callback invoked whenever a new value is received by our monitor. Emits signals based on values changed. """
        if isinstance(value, Disconnected):
            self._connected = False
            self.clear_cache()
            self.connection_state_signal.emit(False)
        else:
            if not self._connected:
                self._connected = True
                self.connection_state_signal.emit(True)
                # Note that there is no way to get the actual write access value from p4p, so defaulting to True for now
                self.write_access_signal.emit(True)

            self._value = value
            has_value_changed_yet = False
            for changed_value in value.changedSet():
                if changed_value == 'value' or changed_value.split('.')[0] == 'value':
                    # NTTable has a changedSet item for each column that has changed
                    # Since we want to send an update on any table change, let's track
                    # if the value item has been updated yet
                    if has_value_changed_yet:
                        continue
                    else:
                        has_value_changed_yet = True
                    
                    if 'NTTable' in value.getID():
                        new_value = value.value.todict()
                        if hasattr(value, 'labels') and 'labels' not in new_value:
                            # Labels are the column headers for the table
                            new_value['labels'] = value.labels
                    elif 'NTEnum' in value.getID():
                        new_value = value.value.index
                        self.enum_strings_signal.emit(tuple(value.value.choices))
                    else:
                        new_value = value.value
                    
                    if self.nttable_data_location:
                        msg = f"Invalid channel... {self.nttable_data_location}"
                        for subfield in self.nttable_data_location:
                            if isinstance(new_value, collections.Container) and not isinstance(new_value, str):     
                                if type(subfield) == str:
                                    try:
                                        new_value = new_value[subfield] 
                                        continue
                                    except (TypeError, IndexError):
                                        logger.debug('Type Error when attempting to use the given key, code will next attempt to convert the key to an int')
                                    except KeyError:
                                        logger.exception(msg)
                                    
                                    try:
                                        new_value = new_value[int(subfield)] 
                                    except ValueError:
                                        logger.exception(msg, exc_info=True)
                            else:
                                logger.exception(msg, exc_info=True)
                                raise ValueError(msg)

                    if new_value is not None:
                        if isinstance(new_value, np.ndarray):
                            if 'NTNDArray' in value.getID():
                                new_value = decompress(value)
                            self.new_value_signal[np.ndarray].emit(new_value)
                        elif isinstance(new_value, np.bool_):
                            self.new_value_signal[np.bool_].emit(new_value)
                        elif isinstance(new_value, list):
                            self.new_value_signal[np.ndarray].emit(np.array(new_value))
                        elif isinstance(new_value, float):
                            self.new_value_signal[float].emit(new_value)
                        elif isinstance(new_value, int):
                            self.new_value_signal[int].emit(new_value)
                        elif isinstance(new_value, str):
                            self.new_value_signal[str].emit(new_value)
                        elif isinstance(new_value, dict):
                            self.new_value_signal[dict].emit(new_value)
                        else:
                            raise ValueError(f'No matching signal for value: {new_value} with type: {type(new_value)}')
                # Sometimes unchanged control variables appear to be returned with value changes, so checking against
                # stored values to avoid sending misleading signals. Will revisit on data plugin changes.
                elif changed_value == 'alarm.severity' and value.alarm.severity != self._severity:
                    self._severity = value.alarm.severity
                    self.new_severity_signal.emit(value.alarm.severity)
                elif changed_value == 'display.precision' and value.display.precision != self._precision:
                    self._precision = value.display.precision
                    self.prec_signal.emit(value.display.precision)
                elif changed_value == 'display.units' and value.display.units != self._units:
                    self._units = value.display.units
                    self.unit_signal.emit(value.display.units)
                elif changed_value == 'control.limitLow' and value.control.limitLow != self._lower_ctrl_limit:
                    self._lower_ctrl_limit = value.control.limitLow
                    self.lower_ctrl_limit_signal.emit(value.control.limitLow)
                elif changed_value == 'control.limitHigh' and value.control.limitHigh != self._upper_ctrl_limit:
                    self._upper_ctrl_limit = value.control.limitHigh
                    self.upper_ctrl_limit_signal.emit(value.control.limitHigh)
                elif changed_value == 'valueAlarm.highAlarmLimit' and \
                        value.valueAlarm.highAlarmLimit != self._upper_alarm_limit:
                    self._upper_alarm_limit = value.valueAlarm.highAlarmLimit
                    self.upper_alarm_limit_signal.emit(value.valueAlarm.highAlarmLimit)
                elif changed_value == 'valueAlarm.lowAlarmLimit' and \
                        value.valueAlarm.lowAlarmLimit != self._lower_alarm_limit:
                    self._lower_alarm_limit = value.valueAlarm.lowAlarmLimit
                    self.lower_alarm_limit_signal.emit(value.valueAlarm.lowAlarmLimit)
                elif changed_value == 'valueAlarm.highWarningLimit' and \
                        value.valueAlarm.highWarningLimit != self._upper_warning_limit:
                    self._upper_warning_limit = value.valueAlarm.highWarningLimit
                    self.upper_warning_limit_signal.emit(value.valueAlarm.highWarningLimit)
                elif changed_value == 'valueAlarm.lowWarningLimit' and \
                        value.valueAlarm.lowWarningLimit != self._lower_warning_limit:
                    self._lower_warning_limit = value.valueAlarm.lowWarningLimit
                    self.lower_warning_limit_signal.emit(value.valueAlarm.lowWarningLimit)
                elif changed_value == 'timeStamp.secondsPastEpoch' and \
                        value.timeStamp.secondsPastEpoch != self._timestamp:
                    self._timestamp = value.timeStamp.secondsPastEpoch
                    self.timestamp_signal.emit(value.timeStamp.secondsPastEpoch)

    @staticmethod
    def convert_epics_nttable(epics_struct):
        """ 
        Converts an epics nttable (passed as a class object p4p.wrapper.Value) to a python dictionary.  
        
        Parameters
        ----------
        epics_struct: 'p4p.wrapper.Value'

        Return
        ------
        result: dict 
        """
        result = {}
        for field in epics_struct.keys():
            value = epics_struct[field]
            if isinstance(value, np.ndarray):
                value = value.tolist()
            elif isinstance(value, p4p.wrapper.Value):
                value = Connection.convert_epics_nttable(value)
            result[field] = value
        return result

    @staticmethod
    def set_value_by_keys(table, keys, new_value):
        """ 
        Saves the passed new_value into the appropriate spot in the given table 
        using the given keys.

        Parameters
        ----------
        table: dict
        keys: list
        new_value: str
        """
        if len(keys) == 1:
            key = keys[0]
            try: 
                table[key] = new_value
            except TypeError:
                table[int(key)] = new_value
        else:
            key = keys[0]

            Connection.set_value_by_keys(table[key], keys[1:], new_value)

    def put_value(self, value):
        """ Write a value to the PV """

        if self.nttable_data_location:
            nttable = Connection.convert_epics_nttable(self._value)
            nttable = nttable["value"]
            Connection.set_value_by_keys(nttable, self.nttable_data_location, value)
            value = {'value': nttable}  

        if is_read_only():
            logger.warning(f'PyDM read-only mode is enabled, could not write value: {value} to {self.address}')
            return

        try:
            P4PPlugin.context.put(self.monitor.name, value)
        except Exception as e:
            logger.error(f"Unable to put value: {value} to channel {self.monitor.name}: {e}")

    def add_listener(self, channel: PyDMChannel):
        """
        Adds a listener to this connection, connecting the appropriate signals/slots to the input PyDMChannel.
        Parameters
        ----------
        channel : PyDMChannel
            The channel that will be listening to any changes from this connection
        """
        super().add_listener(channel)
        if self.monitor is not None and self._connected:
            # Adding a listener to an already connected PV. Manually send the signals indicating the PV is
            # connected, and what the last known values were.
            self.connection_state_signal.emit(True)
            self.write_access_signal.emit(True)
            value_to_send = self._value
            self.clear_cache()
            if value_to_send is not None:
                self.send_new_value(value_to_send)

        if channel.value_signal is not None:
            try:
                channel.value_signal[str].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[int].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[float].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[np.ndarray].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[dict].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass

    def close(self):
        """ Closes out this connection. """
        self.monitor.close()
        super().close()


class P4PPlugin(PyDMPlugin):
    # NOTE: protocol is intentionally "None" to keep this plugin from getting directly imported.
    # If this plugin is chosen as the One True PVA Plugin in pva_plugin.py, the protocol will
    # be properly set before it is used.
    protocol = None
    connection_class = Connection
    context = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if P4PPlugin.context is None:
            # Create the p4p pva context for all connections to use
            context = Context('pva', nt=False)  # Disable automatic value unwrapping
            P4PPlugin.context = context



"""
Loads all the data plugins available at the given PYDM_DATA_PLUGINS_PATH
environment variable and subfolders that follows the *_plugin.py and have
classes that inherits from the pydm.data_plugins.PyDMPlugin class.
"""
import inspect
import logging
import os
from collections import deque
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Type

import entrypoints
from qtpy.QtWidgets import QApplication

from .. import config
from ..utilities import (import_module_by_filename, log_failures, parsed_address)
from .plugin import PyDMPlugin

logger = logging.getLogger(__name__)
plugin_modules: Dict[str, PyDMPlugin] = {}
__read_only = False
global __CONNECTION_QUEUE__
__CONNECTION_QUEUE__ = None
global __DEFER_CONNECTIONS__
__DEFER_CONNECTIONS__ = False
__plugins_initialized = False


@contextmanager
def connection_queue(defer_connections=False):
    global __CONNECTION_QUEUE__
    global __DEFER_CONNECTIONS__
    if __CONNECTION_QUEUE__ is None:
        __CONNECTION_QUEUE__ = deque()
        __DEFER_CONNECTIONS__ = defer_connections
    yield
    if __DEFER_CONNECTIONS__:
        return
    establish_queued_connections()


def establish_queued_connections():
    global __DEFER_CONNECTIONS__
    global __CONNECTION_QUEUE__
    if __CONNECTION_QUEUE__ is None:
        return
    try:
        while (__CONNECTION_QUEUE__ is not None and
               len(__CONNECTION_QUEUE__) > 0):
            channel = __CONNECTION_QUEUE__.popleft()
            establish_connection_immediately(channel)
            QApplication.instance().processEvents()
    except IndexError:
        pass
    finally:
        __CONNECTION_QUEUE__ = None
        __DEFER_CONNECTIONS__ = False


def establish_connection(channel):
    global __CONNECTION_QUEUE__
    if __CONNECTION_QUEUE__ is not None:
        __CONNECTION_QUEUE__.append(channel)
    else:
        establish_connection_immediately(channel)


def establish_connection_immediately(channel):
    plugin = plugin_for_address(channel.address)
    plugin.add_connection(channel)


def plugin_for_address(address: str) -> Optional[PyDMPlugin]:
    """
    Find the correct PyDMPlugin for a channel
    """
    # Check for a configured protocol
    try:
        protocol = parsed_address(address).scheme
    except AttributeError:
        protocol = None 
    
    # Use default protocol
    if protocol is None and config.DEFAULT_PROTOCOL is not None:
        logger.debug("Using default protocol %s for %s",
                     config.DEFAULT_PROTOCOL, address)
        # If no protocol was specified, and the default protocol
        # environment variable is specified, try to use that instead.
        protocol = config.DEFAULT_PROTOCOL

    # Load proper plugin module    
    if protocol:
        initialize_plugins_if_needed()
        try:
            return plugin_modules[str(protocol)]
        except KeyError:
            logger.exception("Could not find protocol for %r", address)
    # Catch all in case of improper plugin specification
    logger.error("Channel {addr} did not specify a valid protocol "
                 "and no default protocol is defined. This channel "
                 "will receive no data. To specify a default protocol, "
                 "set the PYDM_DEFAULT_PROTOCOL environment variable."
                 "".format(addr=address))
    
    return None


def add_plugin(plugin: Type[PyDMPlugin]) -> Optional[PyDMPlugin]:
    """
    Add a PyDM plugin to the global registry of protocol vs. plugins

    Parameters
    ----------
    plugin : PyDMPlugin type
        The class of plugin to instantiate

    Returns
    -------
    plugin : PyDMPlugin, optional
        The instantiated PyDMPlugin. If instantiation failed, will return None.
    """
    # Warn users if we are overwriting a protocol which already has a plugin
    if plugin.protocol in plugin_modules:
        logger.warning(
            "Replacing %s plugin with %s for use with protocol %s",
            plugin,
            plugin_modules[plugin.protocol],
            plugin.protocol,
        )
    try:
        instance = plugin()
    except Exception:
        logger.exception(f"Data plugin: {plugin} failed to load and will not be available for use!")
        return None

    plugin_modules[plugin.protocol] = instance
    return instance


@log_failures(
    logger,
    explanation=(
        "Unable to import plugin file: {args[0]}.  "
        "This plugin will be skipped."
    ),
    include_traceback=True,
)
def _get_plugins_from_source(source_filename: str) -> List[Type[PyDMPlugin]]:
    """
    For a given source filename, find PyDMPlugin classes.

    Parameters
    ----------
    source_filename : str
        The source code filename.

    Returns
    -------
    plugins : list of PyDMPlugin classes
        The plugin classes.
    """
    module = import_module_by_filename(source_filename)
    return list(
        set(
            obj
            for _, obj in inspect.getmembers(module)
            if _is_valid_plugin_class(obj)
        )
    )


def find_plugins_from_path(
    path: str, token: str = config.DATA_PLUGIN_SUFFIX
) -> Generator[Type[PyDMPlugin], None, None]:
    """
    Yield all data plugins found in the provided path.

    Parameters
    ----------
    path : str
        The path to look for plugins.
    token : str, optional
        The suffix that plugin files are expected to have.
    """

    for root, _, files in os.walk(path):
        if root.split(os.path.sep)[-1].startswith("__"):
            continue

        logger.debug("Looking for PyDM Data Plugins at: %s", root)
        for name in files:
            if name.endswith(token):
                yield from _get_plugins_from_source(os.path.join(root, name))


def find_plugins_from_entrypoints(
    key: str = config.ENTRYPOINT_DATA_PLUGIN,
) -> Generator[Type[PyDMPlugin], None, None]:
    """
    Yield all PyDMPlugin classes specified by entrypoints.

    Uses ``entrypoints`` to find packaged external tools in packages that
    configure the ``pydm.data_plugin`` entrypoint.

    Parameters
    ----------
    key : str, optional
        The entrypoint key.
    """
    for entry in entrypoints.get_group_all(key):
        logger.debug("Found data plugin entrypoint: %s", entry.name)
        try:
            plugin_cls = entry.load()
        except Exception as ex:
            logger.exception(
                "Failed to load %s entry %s: %s",
                key, entry.name, ex
            )
            continue

        if not _is_valid_plugin_class(plugin_cls):
            logger.warning(
                "Invalid plugin class specified in entrypoint "
                "%s: %s",
                entry.name, plugin_cls
            )
            continue

        yield plugin_cls


def _is_valid_plugin_class(obj: Any) -> bool:
    """Is the object a data plugin class?"""
    return (
        inspect.isclass(obj)
        and issubclass(obj, PyDMPlugin)
        and obj is not PyDMPlugin
    )


def load_plugins_from_entrypoints(
    key: str = config.ENTRYPOINT_DATA_PLUGIN
) -> Dict[str, PyDMPlugin]:
    """
    Load plugins from file locations that match a specific token

    Parameters
    ----------
    key : str, optional
        The entrypoint key.

    Returns
    -------
    plugins : dict
        Dictionary of plugins add from this folder
    """
    added_plugins = dict()
    for plugin in find_plugins_from_entrypoints(key):
        if not plugin.protocol:
            logger.warning(
                "No protocol specified for data plugin: %s.%s",
                plugin.__module__,
                plugin,
            )
            continue
        added_plugin = add_plugin(plugin)
        if added_plugin is not None:
            added_plugins[plugin.protocol] = added_plugin
    return added_plugins


def load_plugins_from_path(
    locations: List[str],
    token: str = config.DATA_PLUGIN_SUFFIX
) -> Dict[str, PyDMPlugin]:
    """
    Load plugins from file locations that match a specific token

    Parameters
    ----------
    locations : list of str
        List of file locations

    token : str
        Phrase that must match the end of the filename for it to be checked for
        PyDMPlugins

    Returns
    -------
    plugins : dict
        Dictionary of plugins add from this folder
    """
    added_plugins = dict()
    for loc in locations:
        for plugin in find_plugins_from_path(loc, token=token):
            if not plugin.protocol:
                logger.warning(
                    "No protocol specified for data plugin: %s.%s",
                    plugin.__module__,
                    plugin,
                )
                continue

            added_plugin = add_plugin(plugin)
            if added_plugin is not None:
                added_plugins[plugin.protocol] = added_plugin

    return added_plugins


def is_read_only():
    """
    Check whether or not the app is running with the read only flag set.

    Returns
    -------
    bool
        True if read only. False otherwise.
    """
    return __read_only


def set_read_only(read_only):
    """
    Set the read only flag for the data plugins.

    Parameters
    ----------
    read_only : bool
    """
    global __read_only
    __read_only = read_only
    if read_only:
        logger.info("Running PyDM in Read Only mode.")


def initialize_plugins_if_needed():
    global __plugins_initialized

    if __plugins_initialized:
        return

    __plugins_initialized = True

    # Load the data plugins from PYDM_DATA_PLUGINS_PATH
    logger.debug("*"*80)
    logger.debug("* Loading PyDM Data Plugins")
    logger.debug("*"*80)

    path = os.getenv("PYDM_DATA_PLUGINS_PATH", None)
    if path is None:
        locations = []
    else:
        locations = path.split(os.pathsep)

    # Ensure that we first visit the local data_plugins location
    plugin_dir = os.path.dirname(os.path.realpath(__file__))
    locations.insert(0, plugin_dir)

    load_plugins_from_path(locations)
    load_plugins_from_entrypoints(config.ENTRYPOINT_DATA_PLUGIN)



import pytest
import numpy as np
import logging

from qtpy.QtWidgets import QWidget

from pydm.widgets.display_format import  DisplayFormat, parse_value_for_display


# --------------------
# POSITIVE TEST CASES
# --------------------

@pytest.mark.parametrize("value, precision, display_format, widget, expected", [
    (np.array([65, 66], dtype=np.uint8), 1, DisplayFormat.String, QWidget, "AB"),
    (np.array([65, 66, 0, 199], dtype=np.uint8), 1, DisplayFormat.String, QWidget, "AB"),
    ("abc", 0, DisplayFormat.Default, QWidget, "abc"),
    (123, 0, DisplayFormat.Default, QWidget, 123),
    (123.45, 0, DisplayFormat.Default, QWidget, 123.45),
    ("abc", 0, DisplayFormat.String, QWidget, "abc"),
    (123.45, 0, DisplayFormat.Decimal, QWidget, 123.45),
    (3.000e-02, 0, DisplayFormat.Exponential, QWidget, "3e-02"),
    (0x07FF, 0, DisplayFormat.Hex, QWidget, "0x7ff"),
    (0b1101, 0, DisplayFormat.Binary, QWidget, "0b1101"),
])
def test_parse_value_for_display_format(value, precision, display_format, widget, expected):
    """
    Test the correctness of the displayed value according to the specified value type.

    Expectations:
    1. For each value provided, the display format (the string representation of the value) must be as expected
    2. All supported formats are to be tested

    Parameters
    ----------
    value : int, float, hex, bin, str
        The data to be formatted
    precision : int
        The numeric precision to consider during formatting
    display_format : int
        The format type for the provided value
    widget : PyDWidget
        The widget that will display the formatted value. This object can be None
    expected : str
        The expected formatted presentation of the provided value
    """
    parsed_value = parse_value_for_display(
        value, precision, display_format_type=display_format, widget=widget)
    assert(parsed_value == expected)


@pytest.mark.parametrize("value, precision, display_format, widget, expected", [
    (123.45, 1, DisplayFormat.Default, QWidget, 123.45),
    (123.45, 1, DisplayFormat.Decimal, QWidget, 123.45),
    (3.000e-02, 2, DisplayFormat.Exponential, QWidget, "3.00e-02"),
    (3.000e-02, 3, DisplayFormat.Exponential, QWidget, "3.000e-02"),
    (1.234, 3, DisplayFormat.Hex, QWidget, "0x1"),
    (-1.234, 3, DisplayFormat.Hex, QWidget, "-0x2"),
    (1.234, 3, DisplayFormat.Binary, QWidget, "0b1"),
    (-1.234, 3, DisplayFormat.Binary, QWidget, "-0b10")
])
def test_parse_value_for_display_precision(value, precision, display_format, widget, expected):
    """
    Test the correctness of the displayed value according to the specified value precision.

    Expectations:
    The formatted presentation of the displayed value must present the numeric precision as requested, providing a
    precision consideration is applicable for the provided value, for all supported display formats.

    Parameters
    ----------
    value : int, float, hex, bin, str
        The data to be formatted
    precision : int
        The numeric precision to consider during formatting
    display_format : int
        The format type for the provided value
    widget : PyDWidget
        The widget that will display the formatted value. This object can be None
    expected : str
        The expected formatted presentation of the provided value
    """
    assert parse_value_for_display(
        value, precision, display_format_type=display_format, widget=widget) == expected


# --------------------
# NEGATIVE TEST CASES
# ---------------------

@pytest.mark.parametrize("value, precision, display_format, widget, expected", [
    (np.array([-1, -2]), 1, DisplayFormat.String, QWidget, "Could not decode"),
    (np.array([0xfffe, 0xffff]), 1, DisplayFormat.String, QWidget, "Could not decode"),
    ("aaa", 1, DisplayFormat.Exponential, QWidget, "Could not display value 'aaa' using displayFormat 'Exponential'"),
    ("zzz", 2, DisplayFormat.Hex, QWidget, "Could not display value 'zzz' using displayFormat 'Hex'"),
    ("zzz", 3, DisplayFormat.Binary, QWidget, "Could not display value 'zzz' using displayFormat 'Binary'"),
])
def test_parse_value_for_display_precision_incorrect_display_format(
        caplog, value, precision, display_format, widget, expected):
    """
    Test that errors will be output into stderr.

    Parameters
    ----------
    caplog : fixture
        The fixture to capture log outputs
    value : int, float, hex, bin, str
        The incorrect data
    precision : int
        The numeric precision to consider during formatting
    display_format : int
        The format type for the provided value
    widget : PyDWidget
        The widget that will display the formatted value. This object can be None
    expected : str
        The expected formatted presentation of the provided value
    """
    parsed_value = parse_value_for_display(value, precision, display_format_type=display_format, widget=widget)
    if isinstance(value, np.ndarray):
        assert (np.array_equal(value, parsed_value))
    else:
        assert(parsed_value == value)

    # Make sure logging capture the error, and have the correct error message
    for record in caplog.records:
        assert record.levelno == logging.ERROR
    assert expected in caplog.text



# Unit Tests for the Color Map


import pytest

import numpy as np

from ...widgets.colormaps import PyDMColorMap, _magma_data, _inferno_data, _plasma_data, _viridis_data, _jet_data, \
    _monochrome_data, _hot_data, cmaps, magma, inferno, plasma, viridis, jet, monochrome, hot, cmap_names


# --------------------
# POSITIVE TEST CASES
# --------------------

def test_construct():
    """
    Test the construction of the ColorMaps, and the creations of auxiliary helper objects.

    Expecations:
    The default values are assigned to the attributes correctly.
    """

    assert np.array_equal(magma, cmaps[PyDMColorMap.Magma])
    assert np.array_equal(inferno, cmaps[PyDMColorMap.Inferno])
    assert np.array_equal(plasma, cmaps[PyDMColorMap.Plasma])
    assert np.array_equal(viridis, cmaps[PyDMColorMap.Viridis])
    assert np.array_equal(jet, cmaps[PyDMColorMap.Jet])
    assert np.array_equal(monochrome, cmaps[PyDMColorMap.Monochrome])
    assert np.array_equal(hot, cmaps[PyDMColorMap.Hot])

    assert cmap_names[PyDMColorMap.Magma] == "Magma"
    assert cmap_names[PyDMColorMap.Inferno] == "Inferno"
    assert cmap_names[PyDMColorMap.Plasma] == "Plasma"
    assert cmap_names[PyDMColorMap.Viridis] == "Viridis"
    assert cmap_names[PyDMColorMap.Jet] == "Jet"
    assert cmap_names[PyDMColorMap.Monochrome] == "Monochrome"
    assert cmap_names[PyDMColorMap.Hot] == "Hot"



import os
import pytest
import sys
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication, QVBoxLayout
from ...utilities.stylesheet import global_style
from ...widgets.related_display_button import PyDMRelatedDisplayButton

test_ui_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "../test_data", "test.ui")
test_ui_path_with_stylesheet = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "../test_data", "test_emb_style.ui")
test_ui_path_with_relative_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "../test_data", "test_relative_filename_parent.ui")

def test_old_display_filename_property(qtbot):
    # This test is mostly only checking that the related display button
    # doesn't totally explode when the old 'displayFilename' property is used.
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    with pytest.warns(None) as record:
        button.displayFilename = test_ui_path
    assert len(record) >= 1
    assert button.filenames[0] == test_ui_path
    qtbot.addWidget(button)
    button._rebuild_menu()
    qtbot.mouseRelease(button, Qt.LeftButton)
    def check_title():
        assert "Form" in QApplication.instance().main_window.windowTitle()
    qtbot.waitUntil(check_title)

def test_press_with_filename(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    button.filenames = [test_ui_path]
    qtbot.addWidget(button)
    button._rebuild_menu()
    qtbot.mouseRelease(button, Qt.LeftButton)
    def check_title():
        assert "Form" in QApplication.instance().main_window.windowTitle()
    qtbot.waitUntil(check_title)

def test_press_without_filename(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    qtbot.addWidget(button)
    button._rebuild_menu()
    qtbot.mouseRelease(button, Qt.LeftButton)
    qtbot.wait(250)
    assert "Form" not in QApplication.instance().main_window.windowTitle()

def test_no_menu_with_one_file(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    button.filenames = [test_ui_path]
    qtbot.addWidget(button)
    button._rebuild_menu()
    assert button.menu() is None

def test_menu_with_additional_files(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    main_window.set_display_widget(button)
    button.filenames = [test_ui_path, test_ui_path]
    button.titles = ["One", "Two"]
    qtbot.addWidget(button)
    button._rebuild_menu()
    assert button.menu() is not None
    qtbot.mouseRelease(button, Qt.LeftButton)
    qtbot.waitExposed(button.menu())
    qtbot.mouseClick(button.menu(), Qt.LeftButton)
    button.menu().actions()[0].trigger()
    def check_title():
        assert "Form" in QApplication.instance().main_window.windowTitle()
    qtbot.waitUntil(check_title)

def test_menu_goes_away_when_files_removed(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    main_window.set_display_widget(button)
    button.filenames = ["one.ui", "two.ui"]
    button.titles = ["One", "Two"]
    qtbot.addWidget(button)
    button._rebuild_menu()
    assert button.menu() is not None
    button.filenames = []
    button.titles = []
    button._rebuild_menu()
    assert button.menu() is None

def test_menu_goes_away_when_files_all_blank(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    button.filenames = ["", ""]
    button.titles = ["", ""]
    qtbot.addWidget(button)
    button._rebuild_menu()
    assert button.menu() is None

def test_press_with_relative_filename(qtbot):
    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.open(test_ui_path_with_relative_path)
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = main_window.home_widget.relatedDisplayButton
    # Default behavior should be to not follow symlinks (for backwards compat.).
    # Same effect as: button.followSymlinks = False
    qtbot.mouseRelease(button, Qt.LeftButton)
    def check_title():
        assert "Child" in QApplication.instance().main_window.windowTitle()
    qtbot.waitUntil(check_title)

@pytest.mark.skipif(sys.platform == "win32" and sys.version_info < (3, 8), reason="os.path.realpath on Python 3.7 and prior does not resolve symlinks on Windows")
def test_press_with_relative_filename_and_symlink(qtbot, tmp_path):
    symlinked_ui_file = tmp_path / "test_ui_with_relative_path.ui"
    try:
        os.symlink(test_ui_path_with_relative_path, symlinked_ui_file)
    except:
        pytest.skip("Unable to create a symlink for testing purposes.")

    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.open(symlinked_ui_file)
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = main_window.home_widget.relatedDisplayButton
    button.followSymlinks = True
    qtbot.mouseRelease(button, Qt.LeftButton)
    def check_title():
        assert "Child" in QApplication.instance().main_window.windowTitle()
    qtbot.waitUntil(check_title)

def test_no_pydm_app_stylesheet(monkeypatch, qtbot):
    local_is_pydm_app = True

    def is_pydm_app():
        return local_is_pydm_app

    monkeypatch.setattr(
        "pydm.widgets.related_display_button.is_pydm_app",
        is_pydm_app,
    )

    QApplication.instance().make_main_window()
    main_window = QApplication.instance().main_window
    main_window.setWindowTitle("Related Display Button Test")
    qtbot.addWidget(main_window)
    button = PyDMRelatedDisplayButton(parent=main_window)
    qtbot.addWidget(button)

    # In a pydm application, the stylesheet is set on PyDMApplication
    local_is_pydm_app = True
    display1 = button.open_display(test_ui_path)
    assert not display1.styleSheet()

    # In non-pydm applications, we add a stylesheet to the display
    local_is_pydm_app = False
    display2 = button.open_display(test_ui_path)
    assert global_style() in display2.styleSheet()

    # If there was already a stylesheet we need to maintain the original text
    local_is_pydm_app = True
    display3 = button.open_display(test_ui_path_with_stylesheet)
    original_style = display3.styleSheet()
    local_is_pydm_app = False
    display4 = button.open_display(test_ui_path_with_stylesheet)
    assert original_style in display4.styleSheet()
    # And we need to add the global stylesheet too
    assert global_style() in display4.styleSheet()



import pytest
from scipy import constants
from ...utilities import units


@pytest.mark.parametrize("typ, expected", [
    ('cm', 'length'),
    ('non_existent', None)
])
def test_find_unittype(typ, expected):
    tp = units.find_unittype(typ)
    assert (tp == expected)


@pytest.mark.parametrize("unit, expected", [
    ('cm', constants.centi),
    ('non_existent', None)
])
def test_find_unit(unit, expected):
    r = units.find_unit(unit)
    assert (r == expected)


@pytest.mark.parametrize("unit, desired, expected", [
    ('m', 'cm', 1/constants.centi),
    ('m', 'rad', None),
    ('non_existent', 'rad', None)
])
def test_convert(unit, desired, expected):
    r = units.convert(unit, desired)
    assert (r == expected)


@pytest.mark.parametrize("unit, expected", [
    ('V', ['MV', 'kV', 'V', 'mV', 'uV']),
    ('foo', None)
])
def test_find_unit_options(unit, expected):
    opts = units.find_unit_options(unit)
    assert (opts == expected)



import os
import json
import logging
import numpy as np

from datetime import datetime
from typing import Optional

from pydm.widgets.channel import PyDMChannel
from qtpy.QtCore import Slot, QObject, QUrl
from qtpy.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection

logger = logging.getLogger(__name__)


class Connection(PyDMConnection):
    """
    Manages the requests between the archiver data plugin and the archiver appliance itself.
    """

    def __init__(self, channel: PyDMChannel, address: str, protocol: Optional[str] = None,
                 parent: Optional[QObject] = None):
        super().__init__(channel, address, protocol, parent)
        self.add_listener(channel)
        self.address = address
        self.network_manager = QNetworkAccessManager()
        self.network_manager.finished[QNetworkReply].connect(self.data_request_finished)

    def add_listener(self, channel: PyDMChannel) -> None:
        """
        Connects a channel's signal to the slot on this connection so that the channel has a way of requesting
        and receiving data from the archiver.

        Parameters
        ----------
        channel : PyDMChannel
            The channel to connect
        """
        super().add_listener(channel)
        if channel.value_signal is not None:
            channel.value_signal.connect(self.fetch_data)

    def fetch_data(self, from_date: float, to_date: float, processing_command: Optional[str] = None) -> None:
        """
        Fetches data from the Archiver Appliance based on the input parameters.

        Parameters
        ----------
        from_date : float
            Timestamp for the oldest data point to retrieve
        to_date : float
            Timestamp for the newest data point to retrieve
        processing_command : str
            A string that will be added to the URL to request additional processing on the archiver side before
            returning the data such as mean values or optimized. For a full list see:
            https://slacmshankar.github.io/epicsarchiver_docs/userguide.html
            Note: Due to the potential of additional valid options in the future, no validation is
            done on this parameter. It is the responsibility of the caller to ensure it is valid
        """
        if from_date >= to_date:
            logger.error(f"Cannot fetch data for invalid data range, from date={from_date} and to date={to_date}")
            return

        # Archiver expects timestamps to be in utc by default
        from_dt = datetime.utcfromtimestamp(from_date)
        to_dt = datetime.utcfromtimestamp(to_date)

        # Put the dates into the form expected by the archiver in the request url, see here for more details:
        # http://joda-time.sourceforge.net/apidocs/org/joda/time/format/ISODateTimeFormat.html#dateTime()
        from_date_str = from_dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
        to_date_str = to_dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

        base_url = os.getenv("PYDM_ARCHIVER_URL")
        if base_url is None:
            logger.error("Environment variable: PYDM_ARCHIVER_URL must be defined to use the archiver plugin, for "
                         "example: http://lcls-archapp.slac.stanford.edu")
            return

        url_string = f"{base_url}/retrieval/data/getData.json?{self.address}&from={from_date_str}&to={to_date_str}"
        if processing_command:
            url_string = url_string.replace("pv=", "pv=" + processing_command + "(", 1)
            url_string = url_string.replace("&from=", ")&from=", 1)

        request = QNetworkRequest(QUrl(url_string))
        # This get call is non-blocking, can be made in parallel with others, and when the results are ready they
        # will be delivered to the data_request_finished method below via the "finished" signal
        self.network_manager.get(request)

    @Slot(QNetworkReply)
    def data_request_finished(self, reply: QNetworkReply) -> None:
        """
        Invoked when the request to the archiver appliance has been completed and the reply has been returned. Will
        fire off the value signal with a 2D numpy array containing the x-values (timestamps) and y-values (PV data).

        Parameters
        ----------
        reply: The response from the archiver appliance
        """
        if reply.error() == QNetworkReply.NoError and reply.header(QNetworkRequest.ContentTypeHeader) == "application/json":
            bytes_str = reply.readAll()
            data_dict = json.loads(str(bytes_str, 'utf-8'))

            if "pv=optimized" in reply.url().url():  # From a url object to a string
                self._send_optimized_data(data_dict)
            else:
                self._send_raw_data(data_dict)
        else:
            logger.debug(f"Request for data from archiver failed, request url: {reply.url()} retrieved header: "
                         f"{reply.header(QNetworkRequest.ContentTypeHeader)} error: {reply.error()}")
        reply.deleteLater()

    def _send_raw_data(self, data_dict: dict) -> None:
        """
        Sends a numpy array of shape (2, data_length) containing the x-values (timestamps) and y-values (PV data)
        via the new value signal
        """
        data = np.array(([point["secs"] for point in data_dict[0]["data"]],
                         [point["val"] for point in data_dict[0]["data"]]))
        self.new_value_signal[np.ndarray].emit(data)

    def _send_optimized_data(self, data_dict: dict) -> None:
        """
        Sends a numpy array of shape (5, data_length). Index 0 contains the timestamps, index 1 the mean values,
        index 2 the standard deviations, index 3 the minimum values, and index 4 the maximum values.
        """
        pv_data = [point["val"] for point in data_dict[0]["data"]]
        try:
            data = np.array(([point["secs"] for point in data_dict[0]["data"]],
                             [point[0] for point in pv_data],
                             [point[1] for point in pv_data],
                             [point[2] for point in pv_data],
                             [point[3] for point in pv_data]))
        except TypeError:
            # The archiver will fall back to sending raw data if the optimized request is for more data points
            # than are in the bin
            self._send_raw_data(data_dict)
            return

        self.new_value_signal[np.ndarray].emit(data)


class ArchiverPlugin(PyDMPlugin):
    protocol = "archiver"
    connection_class = Connection



import warnings

warnings.warn('pydm.utilities.display_loading was deprecated in favor of pydm.display',
              DeprecationWarning)


def load_py_file(*args, **kwargs):
    from pydm.display import load_py_file
    return load_py_file(*args, **kwargs)



from __future__ import annotations
import functools
import inspect
import logging
import os
import sys
import warnings
from functools import lru_cache
from io import StringIO
from os import path
from string import Template
from typing import Dict, Optional, Tuple

import re
import six
from qtpy import uic
from qtpy.QtWidgets import QApplication, QWidget

from .help_files import HelpWindow
from .utilities import import_module_by_filename, is_pydm_app, macro


class ScreenTarget:
    NEW_PROCESS = 0
    DIALOG = 1
    HOME = 2


logger = logging.getLogger(__file__)


def load_file(file, macros=None, args=None, target=ScreenTarget.NEW_PROCESS):
    """
    Load .ui, .py, or .adl screen file, perform macro substitution, then return
    the resulting QWidget.
    If target is specified, it will properly display the display file.

    Parameters
    ----------
    file : str
        The path to a screen file (.ui, .py, or .adl) to load.
    macros : dict, optional
        A dictionary of macro variables to supply to the
        loaded display subclass.
    args : list, optional
        A list of command-line arguments to pass to the
        loaded display subclass.
    target : int
        One of the ScreenTarget targets. PROCESS is only available when used
        with PyDM Application for now.

    Returns
    -------
    pydm.Display
    """
    if not is_pydm_app() and target == ScreenTarget.NEW_PROCESS:
        logger.warning('New Process is only valid with PyDM Application. ' +
                       'Falling back to ScreenTarget.DIALOG.')
        target = ScreenTarget.DIALOG

    if target == ScreenTarget.NEW_PROCESS:
        # Invoke PyDM to open a new process here.
        app = QApplication.instance()
        app.new_pydm_process(file, macros=macros, command_line_args=args)
        return None

    base, extension = os.path.splitext(file)
    loader = _extension_to_loader.get(extension, load_py_file)
    logger.debug("Loading %s file by way of %s...", file, loader.__name__)
    loaded_display = loader(file, args=args, macros=macros)

    if os.path.exists(base + '.txt'):
        loaded_display.load_help_file(base + '.txt')
    elif os.path.exists(base + '.html'):
        loaded_display.load_help_file(base + '.html')

    if target == ScreenTarget.DIALOG:
        loaded_display.show()

    return loaded_display


@lru_cache()
def _compile_ui_file(uifile: str) -> Tuple[str, str]:
    """
    Compile the ui file using uic and return the result as a string along with the associated class name.
    Caches the result to improve performance when the same ui file is re-used many times within a display.

    Parameters
    ----------
    uifile : str
        The path to a .ui file to compile

    Returns
    -------
    Tuple[str, str] - The first element is the compiled ui file, the second is the name of the class (e.g. Ui_Form)
    """
    code_string = StringIO()
    uic.compileUi(uifile, code_string)
    # Grabs non-whitespace characters between class and the opening parenthesis
    class_name = re.search(r'^class\s*(\S*)\(', code_string.getvalue(), re.MULTILINE).group(1)
    return code_string.getvalue(), class_name


def _load_ui_into_display(uifile, display):
    klass, _ = uic.loadUiType(uifile)

    # Python 2.7 compatibility. More info at the following links:
    # https://github.com/universe-proton/universe-topology/issues/3
    # https://stackoverflow.com/questions/3296993/python-how-to-call-unbound-method-with-other-type-parameter
    retranslateUi = six.get_unbound_function(klass.retranslateUi)
    setupUi = six.get_unbound_function(klass.setupUi)
    # Add retranslateUi to Display class
    display.retranslateUi = functools.partial(retranslateUi, display)
    setupUi(display, display)

    display.ui = display


def _load_compiled_ui_into_display(code_string: str,
                                   class_name: str,
                                   display: Display,
                                   macros: Optional[Dict[str, str]] = None) -> None:
    """
    Takes a ui file which has already been compiled by uic and loads it into the input display.
    Performs macro substitution within the input code_string if any macros supplied are
    found withing the code string.

    Parameters
    ----------
    code_string : str
        The pre-compiled ui file to load
    class_name : str
        The name of the class that methods will be executed on
    display : Display
        The display which the ui file is being loaded into
    macros : Optional[Dict[str, str]]
        Macros to be substituted
    """
    if macros:
        code_string = macro.replace_macros_in_template(Template(code_string), macros).getvalue()
    # Create and grab the class described by the compiled ui file
    ui_globals = {}
    exec(code_string, ui_globals)
    klass = ui_globals[class_name]

    # Add retranslateUi to Display class
    display.retranslateUi = functools.partial(klass.retranslateUi, display)
    klass.setupUi(display, display)

    display.ui = display


def load_ui_file(uifile, macros=None, args=None):
    """
    Load a .ui file, perform macro substitution, then return the resulting QWidget.

    This is an internal method, users will usually want to use `open_file` instead.

    Parameters
    ----------
    uifile : str
        The path to a .ui file to load.
    macros : dict, optional
        A dictionary of macro variables to supply to the file
        to be opened.
    args : list, optional
        This is ignored for UI files.

    Returns
    -------
    QWidget
    """

    display = Display(macros=macros)
    display.load_ui_from_file(uifile, macros)
    return display


def load_adl_file(filename, macros=None, args=None):
    """
    Load an MEDM ADL display with adl2pydm.

    Parameters
    ----------
    filename : str
        The ADL file path.

    macros : dict, optional
        A dictionary of macro variables to supply to the loaded display
        subclass.

    args : any, optional
        Ignored for load_adl_file.
    """
    try:
        import adl2pydm  # noqa: F401
        from adl2pydm import adl_parser, output_handler
    except ImportError:
        raise RuntimeError("Sorry, adl2pydm is not installed.")

    screen = adl_parser.MedmMainWidget(filename)
    buf = screen.getAdlLines(filename)
    screen.parseAdlBuffer(buf)

    writer = output_handler.Widget2Pydm()
    writer.write_ui(screen, None)
    ui_contents = writer.writer.generate_ui_contents()

    d = Display(macros=macros)
    d._loaded_file = filename

    fp = macro.replace_macros_in_template(Template(ui_contents), macros or {})
    _load_ui_into_display(fp, d)
    fp.close()
    return d


def load_py_file(pyfile, args=None, macros=None):
    """
    Load a .py file, performs some sanity checks to try and determine
    if the file actually contains a valid PyDM Display subclass, and if
    the checks pass, create and return an instance.

    This is an internal method, users will usually want to use `open_file` instead.

    Parameters
    ----------
    pyfile : str
        The path to a .py file to load.
    args : list, optional
        A list of command-line arguments to pass to the
        loaded display subclass.
    macros : dict, optional
        A dictionary of macro variables to supply to the
        loaded display subclass.

    Returns
    -------
    pydm.Display
    """
    # Add the intelligence module directory to the python path, so that
    # submodules can be loaded.
    # Eventually, this should go away, and intelligence modules should behave
    # as real python modules.
    module = import_module_by_filename(os.path.abspath(pyfile))

    if hasattr(module, 'intelclass'):
        cls = module.intelclass
        if not issubclass(cls, Display):
            raise ValueError(
                "Invalid class definition at file {}. {} does not inherit from Display. "
                "Nothing to open at this time.".format(
                    pyfile, cls.__name__))
    else:
        classes = [obj for name, obj in inspect.getmembers(module) if
                   inspect.isclass(obj) and issubclass(obj,
                                                       Display) and obj != Display]
        if len(classes) == 0:
            raise ValueError(
                "Invalid File Format. {} has no class inheriting from Display. Nothing to open at this time.".format(
                    pyfile))
        if len(classes) > 1:
            warnings.warn(
                "More than one Display class in file {}. "
                "The first occurrence (in alphabetical order) will be opened: {}".format(
                    pyfile, classes[0].__name__), RuntimeWarning, stacklevel=2)
        cls = classes[0]

    module_params = inspect.signature(cls).parameters

    # Because older versions of Display may not have the args parameter or the macros parameter, we check
    # to see if it does before trying to use them.
    kwargs = {}
    if 'args' in module_params:
        kwargs['args'] = args
    if 'macros' in module_params:
        kwargs['macros'] = macros
    instance = cls(**kwargs)
    instance._loaded_file = pyfile
    return instance


_extension_to_loader = {
    ".ui": load_ui_file,
    ".py": load_py_file,
    ".adl": load_adl_file,
}


class Display(QWidget):

    def __init__(self, parent=None, args=None, macros=None, ui_filename=None):
        super(Display, self).__init__(parent=parent)
        self.ui = None
        self.help_window = None
        self._ui_filename = ui_filename
        self._loaded_file = None
        self._args = args
        self._macros = macros
        self._previous_display = None
        self._next_display = None
        self._local_style = ""
        if ui_filename or self.ui_filename():
            self.load_ui(macros=macros)

    def loaded_file(self):
        return self._loaded_file

    @property
    def previous_display(self):
        return self._previous_display

    @previous_display.setter
    def previous_display(self, display):
        self._previous_display = display

    @property
    def next_display(self):
        return self._next_display

    @next_display.setter
    def next_display(self, display):
        self._next_display = display

    def menu_items(self):
        """ Returns a dictionary where the keys are the names of the menu entries,
        and the values are callables, where the callable is the action performed
        when the menu item is selected.

        Submenus are supported by using a similarly structued dictionary as the value.

        Shortcuts are supported by using a tuple of type (callable, shortcut_string) as the value.

        Users will want to overload this function in their Display subclass to return
        their custom menu.

        Example:

        return {"Action 1": self.action1, "Submenu": {"Action 2" self.action2},
        "Action 3": (self.action3, "Ctrl+A")}

        """
        return {}

    def file_menu_items(self):
        """ Returns a dictionary accepting a protected set of keys corresponding to one or more
        possible default actions in the "File" menu, with the values as callables, where the callable
        is the action performed when the menu item is selected.

        Allowed keys are: ("save", "save_as", "load")

        Shortcuts are supported by using a tuple of type (callable, shortcut_string) as the value.

        Users will want to overload this function in their Display subclass to return
        custom file menu actions.

        Example:

        return {"save": self.save_function, "save_as": self.save_as_function,
        "load": (self.load_function, "Ctrl+L")}

        """
        return {}

    def show_help(self) -> None:
        """ Show the associated help file for this display """
        if self.help_window is not None:
            self.help_window.show()

    def navigate_back(self):
        pass

    def navigate_forward(self):
        pass

    def macros(self):
        if self._macros is None:
            return {}
        return self._macros

    def args(self):
        return self._args

    def ui_filepath(self):
        """ Returns the path to the ui file relative to the file of the class
        calling this function."""
        if not self.ui_filename():
            return None
        path_to_class = sys.modules[self.__module__].__file__
        return path.join(path.dirname(path.realpath(path_to_class)),
                         self.ui_filename())

    def ui_filename(self):
        """ Returns the name of the ui file.  In modern PyDM, it is preferable
        specify this via the ui_filename argument in Display's constructor,
        rather than reimplementing this in Display subclasses."""
        if self._ui_filename is None:
            return None
        else:
            return self._ui_filename

    def load_ui(self, macros=None):
        """ Load and parse the ui file, and make the file's widgets available
        in self.ui.  Called by the initializer."""
        if self.ui:
            return self.ui
        if self.ui_filepath() is not None and self.ui_filepath() != "":
            self.load_ui_from_file(self.ui_filepath(), macros)

    def load_ui_from_file(self, ui_file_path: str, macros: Optional[Dict[str, str]] = None):
        """ Load the ui file from the input path, and make the file's widgets available in self.ui """
        self._loaded_file = ui_file_path
        code_string, class_name = _compile_ui_file(ui_file_path)
        _load_compiled_ui_into_display(code_string, class_name, self, macros)

    def load_help_file(self, file_path: str) -> None:
        """ Loads the input help file into a window for display """
        self.help_window = HelpWindow(file_path)

    def setStyleSheet(self, new_stylesheet):
        # Handle the case where the widget's styleSheet property contains a filename, rather than a stylesheet.
        possible_stylesheet_filename = os.path.expanduser(os.path.expandvars(new_stylesheet))
        logger.debug("Calling Display.setStyleSheet, new_stylesheet is %s", possible_stylesheet_filename)
        stylesheet_filename = None
        try:
            # First, check if the file is already an absolute path.
            if os.path.isfile(possible_stylesheet_filename):
                stylesheet_filename = possible_stylesheet_filename
            # Second, check if the css file is specified relative to the display file.
            else:
                rel_path = os.path.join(os.path.dirname(os.path.abspath(self._loaded_file)), possible_stylesheet_filename)
                if os.path.isfile(rel_path):
                    stylesheet_filename = rel_path
        except Exception as e:
            logger.debug("Exception while checking if stylesheet is a filename: %s", e)
            pass
        self._local_style = new_stylesheet
        if stylesheet_filename is not None:
            logger.debug("styleSheet property contains a filename, loading %s", stylesheet_filename)
            with open(stylesheet_filename) as f:
                self._local_style = f.read()
        logger.debug("Setting stylesheet to: %s", self._local_style)
        super(Display, self).setStyleSheet(self._local_style)
